#!/usr/bin/env python

from collections import Iterator, Mapping, Sequence

class RowIterator(Iterator):
    """
    Iterator over row proxies.
    """
    def __init__(self, table):
        self._table = table
        self._i = 0

    def next(self):
        try:
            self._table._data[self._i]
        except IndexError:
            raise StopIteration
        
        i = self._i
        self._i += 1

        return Row(self._table, i)

class RowSequence(Sequence):
    """
    Proxy access to rows by index.
    """
    def __init__(self, table):
        self._table = table

    def __getitem__(self, i):
        if isinstance(i, slice):
            indices = xrange(*i.indices(len(self)))

            return [Row(self._table, row) for row in indices]

        self._table._data[i]

        return Row(self._table, i)

    def __len__(self):
        return len(self._table._data)

class CellIterator(Iterator):
    """
    Iterator over row cells.
    """
    def __init__(self, row):
        self._row = row
        self._i = 0

    def next(self):
        try:
            v = self._row._table._data[self._row._i][self._i]
        except IndexError:
            raise StopIteration
        
        self._i += 1

        return v 

class Row(Mapping):
    """
    Proxy to row data.
    """
    def __init__(self, table, i):
        self._table = table
        self._i = i

    def __getitem__(self, k):
        j = self._table._column_names.index(k)

        return self._table._data[self._i][j]

    def __len__(self):
        return len(self._table._data[self._i])

    def __iter__(self):
        return CellIterator(self)

    def __eq__(self, other):
        return self._table._data[self._i] == other



#!/usr/bin/env python

from journalism.columns import TextColumn, IntColumn, FloatColumn
from journalism.exceptions import *
from journalism.table import Table

def save():
    raise NotImplementedError



#!/usr/bin/env python

class NullComputationError(Exception):
    """
    Exception raised if an illogical computation is
    attempted on a Column containing nulls.
    """
    pass

class UnsupportedOperationError(Exception):
    """
    Exception raised when an operation is applied
    to an invalid column type.
    """
    def __init__(self, operation, column):
        self.operation = operation
        self.column = column

    def __unicode__(self):
        return '`%s` is not a supported operation for %s' % (self.operation, type(self.column))

    def __str__(self):
        return str(self.__unicode__())

class ColumnValidationError(Exception):
    """
    Exception raised in a column value can not be
    validated.
    """
    def __init__(self, value, column):
        self.value = value
        self.column = column

    def __unicode__(self):
        return '`%s` is not a valid value for %s' % (self.value, type(self.column))

    def __str__(self):
        return str(self.__unicode__())




#!/usr/bin/env python

"""
This module contains the Table object.
"""

from journalism.columns import ColumnMapping
from journalism.exceptions import UnsupportedOperationError
from journalism.rows import RowSequence

def transpose(data):
    """
    Utility function for transposing a 2D array of data.
    """
    return zip(*data)

class Table(object):
    """
    A group of columns with names.

    TODO: dedup column names
    """
    def __init__(self, rows, column_types=[], column_names=[], cast=False, validate=False):
        """
        Create a table from rows of data.

        TODO: validate column_types are all subclasses of Column.
        """
        if not column_names:
            column_names = [unicode(d) for d in rows.pop(0)]

        self._data = rows
        self._column_types = column_types
        self._column_names = column_names

        self.columns = ColumnMapping(self)
        self.rows = RowSequence(self)

        if cast:
            data_columns = []

            for column in self.columns:
                data_columns.append(column._cast())
            
            self._data = transpose(data_columns)

        if validate:
            for column in self.columns:
                column.validate()

    def _fork(self, new_data, column_types=[], column_names=[]):
        """
        Create a new table using the metadata from this one.
        Used internally by functions like :meth:`sort_by`.
        """
        if not column_types:
            column_types = self._column_types

        if not column_names:
            column_names = self._column_names

        return Table(new_data, column_types, column_names)

    def sort_by(self, column_name, cmp=None, reverse=False):
        """
        Sort this table by the specified column.

        Returns a new :class:`Table`.
        """
        i = self._column_names.index(column_name)

        data = sorted(self._data, cmp=cmp, key=lambda r: r[i], reverse=reverse)

        return self._fork(data)

    def filter(self, column_name, include=[]):
        """
        Filter a to only those rows where the column is in the
        include list.

        Returns a new :class:`Table`.
        """
        i = self._column_names.index(column_name)

        rows = [row for row in self._data if row[i] in include]

        return Table(rows, self._column_types, self._column_names)

    def reject(self, column_name, exclude=[]):
        """
        Filter a to only those rows where the column is not in the
        exclude list.

        Returns a new :class:`Table`.
        """
        i = self._column_names.index(column_name)

        rows = [row for row in self._data if row[i] not in exclude]

        return Table(rows, self._column_types, self._column_names)

    def where(self, column_name, cmp):
        """
        Filter the data to only those rows where the specified column
        passes a truth test.

        Returns a new :class:`Table`.
        """
        i = self._column_names.index(column_name)

        rows = filter(lambda r: cmp(r[i]), self._data)

        return Table(rows, self._column_types, self._column_names)

    def aggregate(self, group_by, operations=[]):
        """
        Aggregate data by a specified group_by column.

        Operations is a dict of column names and operation names.

        Returns a new :class:`Table`.
        """
        i = self._column_names.index(group_by)

        groups = {}

        for row in self._data:
            group_name = row[i]

            if group_name not in groups:
                groups[group_name] = []

            groups[group_name].append(row)

        output = []

        column_types = [self._column_types[i]]
        column_names = [group_by]

        for op_column in [op[0] for op in operations]:
            column_types.append(self._column_names.index(op_column))
            column_names.append(op_column)

        for name, group_rows in groups.items():
            group_table = Table(group_rows, self._column_types, self._column_names) 
            new_row = [name]

            for op_column, operation in operations:
                c = group_table.columns[op_column]
                
                try:
                    op = getattr(c, operation)
                except AttributeError:
                    raise UnsupportedOperationError(operation, c)

                new_row.append(op())

            output.append(new_row)
        
        return Table(output, column_types, column_names) 




#!/usr/bin/env python

from collections import Iterator, Mapping, Sequence
import copy
from functools import wraps
import math

from journalism.exceptions import ColumnValidationError, NullComputationError

class ColumnIterator(Iterator):
    """
    Iterator over column proxies.
    """
    def __init__(self, table):
        self._table = table
        self._i = 0

    def next(self):
        try:
            v = self._table._column_names[self._i]
        except IndexError:
            raise StopIteration

        column_type = self._table._column_types[self._i]

        self._i += 1
        
        return column_type(self._table, v)

class ColumnMapping(Mapping):
    """
    Proxy access to columns by name.
    """
    def __init__(self, table):
        self._table = table

    def __getitem__(self, k):
        if k not in self._table._column_names:
            raise KeyError

        i = self._table._column_names.index(k)
        column_type = self._table._column_types[i]

        return column_type(self._table, k)

    def __iter__(self):
        return ColumnIterator(self._table)

    def __len__(self):
        return len(self._table._column_names)

def no_null_computations(func):
    """
    Function decorator that prevents illogical computations
    on columns containing nulls.
    """
    @wraps(func)
    def check(l, *args, **kwargs):
        if l.has_nulls():
            raise NullComputationError

        return func(l)

    return check

class Column(Sequence):
    """
    Proxy access to column data.
    """
    def __init__(self, table, k):
        self._table = table
        self._k = k

    def _data(self):
        # TODO: memoize?
        i = self._table._column_names.index(self._k)

        return [r[i] for r in self._table._data]

    def _data_without_nulls(self):
        # TODO: memoize?
        return [d for d in self._data() if d is not None]

    def _data_sorted(self):
        # TODO: memoize?
        return sorted(self._data())

    def __getitem__(self, j):
        return self._data()[j]

    def __len__(self):
        return len(self._data()) 

    def __eq__(self, other):
        return self._data() == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def validate(self):
        """
        Verify values in this column are of an appopriate type.
        """
        raise NotImplementedError

    def _cast(self):
        """
        Cast values in this column to an appropriate type, if possible.
        """
        raise NotImplementedError

    def has_nulls(self):
        """
        Returns True if this column contains null values.
        """
        return None in self._data()

    def map(self, func, new_column_type=None, new_column_name=None):
        """
        Apply an arbitrary function to a column of data and
        optionally change it's type and/or name.

        Returns a new :class:`journalism.table.Table`.
        """
        i = self._table._column_names.index(self._k)
        
        data = copy.deepcopy(self._table._data)
        column_types = copy.deepcopy(self._table._column_types)
        column_names = copy.deepcopy(self._table._column_names)

        for row in data:
            row[i] = func(row[i])

        if new_column_type:
            column_types[i] = new_column_type

        if new_column_name:
            column_names[i] = new_column_name

        return self._table._fork(data, column_types, column_names) 

class TextColumn(Column):
    """
    A column containing unicode/string data.
    """
    def validate(self):
        """
        Verify all values in this column are string/unicode or null.

        Will raise :exc:`journalism.exceptions.ColumnValidationError`
        if validation fails.
        """
        for d in self._data():
            if not isinstance(d, basestring) and d is not None:
                raise ColumnValidationError(d, self)

    def _cast(self):
        """
        Cast values to unicode.
        """
        casted = []

        for d in self._data():
            if d == '':
                casted.append(None)
            else:
                casted.append(unicode(d))

        return casted

class NumberColumn(Column):
    """
    A column containing numeric data.
    
    Base class for :class:`IntColumn` and :class:`FloatColumn`.
    """
    def sum(self):
        """
        Implemented in subclasses to take advantage of floating
        point precision.
        """
        raise NotImplementedError

    def min(self):
        """
        Compute the minimum value of this column.
        """
        return min(self._data_without_nulls())

    def max(self):
        """
        Compute the maximum value of this column.
        """
        return max(self._data_without_nulls())

    @no_null_computations
    def mean(self):
        """
        Compute the mean value of this column.

        Will raise :exc:`journalism.exceptions.NullComputationError` if this column contains nulls.
        """
        return float(self.sum() / len(self))

    @no_null_computations
    def median(self):
        """
        Compute the median value of this column.

        Will raise :exc:`journalism.exceptions.NullComputationError` if this column contains nulls.
        """
        data = self._data_sorted()
        length = len(data)

        if length % 2 == 1:
            return data[((length + 1) / 2) - 1]
        else:
            a = data[(length / 2) - 1]
            b = data[length / 2]

        return (float(a + b)) / 2  

    @no_null_computations
    def mode(self):
        """
        TODO: Compute the mode of this column.

        Will raise :exc:`journalism.exceptions.NullComputationError` if this column contains nulls.
        """
        raise NotImplementedError

    @no_null_computations
    def stdev(self):
        """
        Compute the standard of deviation of this column.

        Will raise :exc:`journalism.exceptions.NullComputationError` if this column contains nulls.
        """
        data = self._data()

        return math.sqrt(sum(math.pow(v - self.mean(), 2) for v in data) / len(data))

class IntColumn(NumberColumn):
    """
    A column containing integer data.
    """
    def validate(self):
        """
        Verify all values in this column are int or null.

        Will raise :exc:`journalism.exceptions.ColumnValidationError`
        if validation fails.
        """
        for d in self._data():
            if not isinstance(d, int) and d is not None:
                raise ColumnValidationError(d, self)

    def _cast(self):
        """
        Cast values in this column to integer.
        """
        casted = []

        for d in self._data():
            if isinstance(d, basestring):
                d = d.replace(',' ,'').strip()

            if d == '' or d is None:
                casted.append(None)
            else:
                casted.append(int(d))

        return casted

    def sum(self):
        """
        Compute the sum of this column.
        """
        return sum(self._data_without_nulls())

class FloatColumn(NumberColumn):
    """
    A column containing float data.
    """
    def validate(self):
        """
        Verify all values in this column are float or null.

        Will raise :exc:`journalism.exceptions.ColumnValidationError`
        if validation fails.
        """
        for d in self._data():
            if not isinstance(d, float) and d is not None:
                raise ColumnValidationError(d, self)

    def _cast(self):
        """
        Cast values in this column to float.
        """
        casted = []

        for d in self._data():
            if isinstance(d, basestring):
                d = d.replace(',' ,'').strip()

            if d == '' or d is None:
                casted.append(None)
            else:
                casted.append(float(d))

        return casted

    def sum(self):
        """
        Compute the sum of this column using :func:`math.fsum` for precision.
        """
        return math.fsum(self._data_without_nulls())




#!/usr/bin/env python

from setuptools import setup

setup(
    name='journalism',
    version='0.1.0',
    description='',
    long_description=open('README').read(),
    author='Christopher Groskopf',
    author_email='staringmonkey@gmail.com',
    url='journalism takes the horror out of basic data analysis and manipulation.',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=[
        'journalism',
    ],
    install_requires = [
    ]
)
